pub(crate) mod proto {
  include!(concat!(env!("OUT_DIR"), "/computer.rs"));
}

use std::sync::{mpsc, Arc};

use jni::{
  objects::{JByteArray, JByteBuffer, JObject, JObjectArray, JValue},
  signature::{Primitive, ReturnType},
  sys::jlong,
  JNIEnv,
};
use prost::Message;
use proto::component_init::Data;
use zerocopy::{little_endian::U64, FromZeros};

use crate::{
  align_up_to_page,
  components::{
    core::{CoreComponent, Registers},
    debug_chat::DebugChat,
    AllComponent,
  },
  core::{spawn_core_thread, Core, CoreMessage},
  device_bus::{DeviceBus, ROM_START},
  unsync_cell::UnsyncCell,
};

pub struct Computer {
  cores: Vec<mpsc::Sender<CoreMessage>>,
}

fn try_get_object<'local>(
  env: &mut JNIEnv<'local>,
  init_objects: &JObjectArray<'local>,
  index: i32,
) -> jni::errors::Result<Option<JObject<'local>>> {
  if index < 0 {
    return Ok(None);
  }

  env.get_object_array_element(&init_objects, index).map(Some)
}

impl Computer {
  #[unsafe(export_name = "Java_ca_sanae_golemcomputers_computer_Computer_new")]
  unsafe extern "C" fn new<'local>(
    mut env: JNIEnv<'local>,
    java_computer: JObject<'local>,
    init_data: JByteArray<'local>,
    init_objects: JObjectArray<'local>,
  ) {
    let init_data = env.convert_byte_array(&init_data).expect("unable to convert byte array to vec");
    let init = proto::ComputerInit::decode(init_data.as_slice()).expect("failed to decode init data from java!");
    let ram = UnsyncCell::new_zeroed_box(init.primary_ram_pages as usize * 0x1000);

    let rom = try_get_object(&mut env, &init_objects, init.primary_rom_index)
      .expect("failed to get rom from java")
      .map(|rom| {
        if !env.is_instance_of(&rom, "java/nio/ByteBuffer").expect("failed to perform instanceOf") {
          panic!("rom was not a byte buffer")
        }
        let rom_byte_buf = JByteBuffer::from(rom);
        let capacity = env.get_direct_buffer_capacity(&rom_byte_buf).expect("unable to extract capacity from rom");
        let rom = UnsyncCell::new_zeroed_box(align_up_to_page(capacity));
        rom.update_from_jni(&mut env, &rom_byte_buf).unwrap();
        rom
      })
      .expect("no rom was provided");

    let mut core_init_datas = Vec::new();
    let components = init
      .components
      .into_iter()
      .map(|component| match component.data.unwrap() {
        Data::Core(init) => {
          let registers = Arc::new(UnsyncCell::new(Registers {
            reset_address: U64::new(ROM_START),
            ..Registers::new_zeroed()
          }));
          let (sender, receiver) = mpsc::channel();
          core_init_datas.push((init, registers.clone(), sender.clone(), receiver));
          AllComponent::Core(CoreComponent::new(registers, sender))
        }
        Data::Chat(init) => {
          let component = try_get_object(&mut env, &init_objects, init.java_index)
            .expect("failed to get chat component from java")
            .expect("chat component object was not provided");

          let component = env.new_global_ref(component).unwrap();

          AllComponent::DebugChat(DebugChat::new(component))
        }
      })
      .collect();

    let device_bus = Arc::new(DeviceBus::new(ram, rom, components));

    let cores = core_init_datas
      .into_iter()
      .map(|(init, registers, sender, receiver)| {
        spawn_core_thread(
          Core::new(registers, device_bus.clone(), init.instructions_per_tick),
          env.get_java_vm().unwrap(),
          receiver,
        );
        sender
      })
      .collect();

    let computer = Box::new(Self { cores });

    env
      .set_field(
        &java_computer,
        "address",
        "J",
        JValue::Long(Box::into_raw(computer) as jlong),
      )
      .unwrap();
  }

  fn from_jni<'local>(env: &mut JNIEnv<'local>, computer: JObject<'local>) -> *mut Self {
    env
      .get_field_unchecked(
        &computer,
        (env.get_object_class(&computer).unwrap(), "address", "J"),
        ReturnType::Primitive(Primitive::Long),
      )
      .unwrap()
      .j()
      .unwrap() as *mut _
  }

  #[unsafe(export_name = "Java_ca_sanae_golemcomputers_computer_Computer_free")]
  unsafe extern "C" fn free<'local>(mut env: JNIEnv<'local>, computer: JObject<'local>) {
    let computer = Self::from_jni(&mut env, computer);
    // Safety: the responsibility of asserting the pointer is valid is on the java caller
    drop(unsafe { Box::from_raw(computer) })
  }

  #[unsafe(export_name = "Java_ca_sanae_golemcomputers_computer_Computer_resetCore")]
  unsafe extern "C" fn reset_core<'local>(mut env: JNIEnv<'local>, computer: JObject<'local>, core: u32, wake: bool) {
    // Safety: the responsibility of asserting the pointer is valid is on the java caller
    let computer = unsafe { &*Self::from_jni(&mut env, computer) };
    let core = computer.cores.get(core as usize).unwrap();
    core.send(CoreMessage::Reset).unwrap();
    if wake {
      core.send(CoreMessage::Wake).unwrap();
    }
  }

  #[unsafe(export_name = "Java_ca_sanae_golemcomputers_computer_Computer_tick")]
  unsafe extern "C" fn tick<'local>(mut env: JNIEnv<'local>, computer: JObject<'local>) {
    // Safety: the responsibility of asserting the pointer is valid is on the java caller
    let computer = unsafe { &*Self::from_jni(&mut env, computer) };
    for core in &computer.cores {
      core.send(CoreMessage::Tick).unwrap();
    }
  }
}
