use std::sync::{mpsc, Arc};

use disarm64::decoder;
use jni::{AttachGuard, JNIEnv, JavaVM};
use num_derive::FromPrimitive;
use num_traits::FromPrimitive;
use unicorn_engine::{Arch, HookType, Mode, Unicorn};
use zerocopy::IntoBytes;

use crate::{components::core::Registers, device_bus::DeviceBus, unsync_cell::UnsyncCell};

pub struct Core {
  registers: Arc<UnsyncCell<Registers>>,
  bus: Arc<DeviceBus>,
  instructions_per_tick: u32,
}

impl Core {
  pub fn new(registers: Arc<UnsyncCell<Registers>>, bus: Arc<DeviceBus>, instructions_per_tick: u32) -> Self {
    Self {
      registers,
      bus,
      instructions_per_tick,
    }
  }
}

pub struct CoreContext<'a> {
  core: Core,
  env: AttachGuard<'a>,
  is_sleeping: bool,
}

impl<'a> CoreContext<'a> {
  pub fn get_env(&mut self) -> &mut JNIEnv<'a> {
    &mut self.env
  }
}

pub enum CoreMessage {
  Tick,
  Reset,
  Wake,
  Sleep,
}

pub fn spawn_core_thread(core: Core, jvm: JavaVM, message_receiver: mpsc::Receiver<CoreMessage>) {
  std::thread::spawn(|| core_main(core, jvm, message_receiver));
}

fn core_main(core: Core, jvm: JavaVM, message_receiver: mpsc::Receiver<CoreMessage>) {
  let env = jvm.attach_current_thread().unwrap();
  let mut uc = Unicorn::new_with_data(
    Arch::ARM64,
    Mode::LITTLE_ENDIAN,
    CoreContext {
      core,
      env,
      is_sleeping: true,
    },
  )
  .unwrap();

  let bus = uc.get_data().core.bus.clone();
  bus.map(&mut uc);
  uc.add_intr_hook(interrupt_handler).unwrap();
  uc
  .add_mem_hook(
    HookType::MEM_INVALID,
    u64::MIN,
    u64::MAX,
    |vcpu, mem_type, address, _, value| {
      eprintln!("memory exception: {mem_type:?} at {address:X} (value: {value:X})");

      vcpu.emu_stop().unwrap();
      true
    },
  )
  .unwrap();

  let try_fetch_message = |stopped| {
    if stopped {
      message_receiver.recv().map(Some).map_err(|_| ())
    } else {
      Ok(message_receiver.try_recv().ok())
    }
  };

  let mut instructions_left = 0;
  loop {
    while let Some(message) = try_fetch_message(uc.get_data().is_sleeping || instructions_left == 0).transpose() {
      let Ok(message) = message else { return };

      match message {
        CoreMessage::Tick => {
          if uc.get_data().is_sleeping {
            continue;
          }
          println!("ticking");
          instructions_left = uc.get_data().core.instructions_per_tick as usize;
        }
        CoreMessage::Reset => {
          let reset_address = uc.get_data().core.registers.read().reset_address.get();
          println!("reset! address: {reset_address:08X}");
          uc.set_pc(reset_address).expect("error while executing cpu")
        }
        CoreMessage::Wake => uc.get_data_mut().is_sleeping = false,
        CoreMessage::Sleep => uc.get_data_mut().is_sleeping = true,
      }
    }
    println!("executing");
    uc.emu_start(
      uc.pc_read().expect("error while executing cpu"),
      0,
      0,
      instructions_left.min(1000),
    )
    .expect("error while executing cpu");
    instructions_left = instructions_left.saturating_sub(1000);
  }
}

#[derive(Debug, FromPrimitive)]
enum Exception {
  UndefinedInstruction = 1,
  SupervisorCall = 2,
  PrefetchAbort = 3,
  DataAbort = 4,
  Interrupt = 5,
  FastInterrupt = 6,
  Breakpoint = 7,
  ExceptionExit = 8,
  KernelTrap = 9,
  HypervisorCall = 11,
  HypervisorTrap = 12,
  SecureMonitorCall = 13,
  VirtualInterrupt = 14,
  VirtualFastInterrupt = 15,
  Semihost = 16,
  Nocp = 17,
  Invstate = 18,
  Stkof = 19,
  Lazyfp = 20,
  Lserr = 21,
  Unaligned = 22,
}

fn interrupt_handler(uc: &mut Unicorn<CoreContext>, kind: u32) {
  let exception = Exception::from_u32(kind).unwrap();

  let pc = uc.pc_read().unwrap();
  println!("exception: {exception:?}");
  let mut opcode = 0u32;
  if let Ok(_) = uc.mem_read(pc, opcode.as_mut_bytes()) {
    if let Some(opcode) = decoder::decode(opcode) {
      let mut inst = String::new();
      disarm64::format_insn::format_insn_pc(pc, &mut inst, &opcode).unwrap();
      println!("pc {pc:08X} {inst}");
    } else {
      println!("pc {pc:08X} - op parse failed {opcode:08X}");
    }
  } else {
    println!("pc {pc:08X} - can't read");
  }

  use unicorn_engine::RegisterARM64::*;
  println!("X0: {:X}", uc.reg_read(X0).unwrap());
  println!("X1: {:X}", uc.reg_read(X1).unwrap());
  println!("X2: {:X}", uc.reg_read(X2).unwrap());

  // if an exception occurs, we don't want to execute anymore
  uc.get_data_mut().is_sleeping = true;
}
