use std::{num::NonZeroUsize, sync::Arc};

use bitfield_struct::bitfield;
use nodit::{interval::ie, Interval, NoditMap, OverlapError};
use unicorn_engine::{Permission, Unicorn};
use zerocopy::{little_endian::U32, FromBytes, FromZeros, Immutable, IntoBytes, KnownLayout};

use crate::{align_up_to_page, components::AllComponent, core::CoreContext, unsync_cell::UnsyncCell};

pub const COMPONENT_TABLE_START: u64 = 0x1000_0000;
pub const COMPONENTS_START: u64 = 0x2000_0000;
pub const ROM_START: u64 = 0x4000_0000;
pub const RAM_START: u64 = 0x8000_0000;

enum Device {
  Rom(Box<UnsyncCell<[u8]>>),
  Ram(Box<UnsyncCell<[u8]>>),
  ComponentTable,
  Component(AllComponent),
}

#[bitfield(u32)]
struct ComponentEntryHeader {
  #[bits(31)]
  _padding: u32,
  filled: bool,
}

#[derive(Clone, Copy, FromBytes, IntoBytes, Immutable, KnownLayout)]
#[repr(C)]
struct ComponentTableEntry {
  header: U32,
  address: U32,
  name: [u8; 0x30],
}

struct ComponentTable(Box<[u8]>);

pub struct DeviceBus {
  devices: NoditMap<u64, Interval<u64>, Device>,
  component_table: ComponentTable,
}

fn assert_non_overlapping(result: Result<(), OverlapError<Device>>) {
  if result.is_err() {
    panic!("insert overlapped!")
  }
}

impl DeviceBus {
  pub fn new(ram: Box<UnsyncCell<[u8]>>, rom: Box<UnsyncCell<[u8]>>, components: Vec<AllComponent>) -> Self {
    let mut devices = NoditMap::new();

    assert_non_overlapping(devices.insert_strict(ie(ROM_START, ROM_START + rom.len() as u64), Device::Rom(rom)));
    assert_non_overlapping(devices.insert_strict(ie(RAM_START, RAM_START + ram.len() as u64), Device::Ram(ram)));

    let mut component_table = {
      let mut table = Box::new_uninit_slice(align_up_to_page(components.len() * size_of::<ComponentTableEntry>()));
      table.zero();
      unsafe { table.assume_init() }
    };
    assert_non_overlapping(devices.insert_strict(
      ie(
        COMPONENT_TABLE_START,
        COMPONENT_TABLE_START + component_table.len() as u64,
      ),
      Device::ComponentTable,
    ));

    {
      let component_table = <[ComponentTableEntry]>::mut_from_bytes_with_elems(
        &mut component_table[..components.len() * size_of::<ComponentTableEntry>()],
        components.len(),
      )
      .unwrap();

      let mut component_offset = COMPONENTS_START;
      for (index, component) in components.into_iter().enumerate() {
        let start = component_offset;
        let end = align_up_to_page(component_offset + component.memory_map_size());
        println!("component offset: {start:X}..{end:X} and {}", u32::MAX);
        assert!(end < ROM_START);

        component_offset = end;
        component_table[index] = ComponentTableEntry {
          header: U32::new(ComponentEntryHeader::new().with_filled(true).into_bits()),
          address: U32::new(start as u32),
          name: component.name(),
        };

        assert_non_overlapping(devices.insert_strict(ie(start, end), Device::Component(component)));
      }
    }

    Self {
      devices,
      component_table: ComponentTable(component_table),
    }
  }

  pub fn map(self: Arc<Self>, uc: &mut Unicorn<CoreContext>) {
    for (interval, device) in self.devices.iter() {
      let (address, size) = (interval.start(), (interval.end() - interval.start()) as usize + 1);
      println!("address, size {address:X}, {size:X} {:X?}", interval);
      match device {
        Device::Rom(memory) => unsafe {
          uc.mem_map_ptr(
            address,
            size,
            Permission::READ | Permission::WRITE,
            memory.as_mut_slice_ptr() as _,
          )
          .unwrap();
        },
        Device::Ram(memory) => unsafe {
          uc.mem_map_ptr(
            address,
            size,
            Permission::READ | Permission::WRITE | Permission::EXEC,
            memory.as_mut_slice_ptr() as _,
          )
          .unwrap();
        },
        Device::ComponentTable => unsafe {
          println!("component table range {address:X} + {size:X}");
          uc.mem_map_ptr(address, size, Permission::READ, self.component_table.0.as_ptr() as _).unwrap()
        },
        Device::Component(_) => {
          fn calculate_size(component: &AllComponent, offset: u64, size: usize) -> Option<NonZeroUsize> {
            if offset >= component.memory_map_size() {
              return None;
            }

            let offset = offset as usize;
            NonZeroUsize::new(usize::min(offset + size, component.memory_map_size() as usize) - offset)
          }
          let bus_read = self.clone();
          let bus_write = self.clone();
          uc.mmio_map(
            address,
            size,
            Some(move |uc: &mut Unicorn<CoreContext>, offset: u64, size: usize| {
              let Device::Component(component) = bus_read.devices.get_at_point(address).unwrap() else {
                unreachable!("device was not a component")
              };

              let Some(size) = calculate_size(component, offset, size).map(NonZeroUsize::get) else {
                return 0;
              };

              let mut data = 0u64;
              component.read(uc.get_data_mut().get_env(), offset, &mut data.as_mut_bytes()[..size]);
              data
            }),
            Some(
              move |uc: &mut Unicorn<CoreContext>, offset: u64, size: usize, data: u64| {
                let Device::Component(component) = bus_write.devices.get_at_point(address).unwrap() else {
                  unreachable!("device was not a component")
                };

                let Some(size) = calculate_size(component, offset, size).map(NonZeroUsize::get) else {
                  return;
                };

                component.write(uc.get_data_mut().get_env(), offset, &data.to_le_bytes()[..size]);
              },
            ),
          )
          .unwrap()
        }
      }
    }
  }
}
