#![deny(unsafe_op_in_unsafe_fn)]
#![feature(sync_unsafe_cell)]

use std::ops::Range;

use jni::{
  sys::{jint, JNI_VERSION_1_8},
  JNIEnv,
};
use num_traits::PrimInt;

pub mod components;
mod computer;
pub mod core;
mod device_bus;
pub mod unsync_cell;

#[macro_export]
macro_rules! assert_positive {
  ($env: expr, $value: expr) => {{
    let value = $value;
    if value < 0 {
      ($env)
        .throw(format!(
          "input {} from java cannot be negative! got {value}",
          stringify!($value)
        ))
        .unwrap();
      return;
    }
    value as _
  }};
  ($env: expr, $value: expr, $err: expr) => {{
    let value = $value;
    if value < 0 {
      ($env)
        .throw(format!(
          "input {} from java cannot be negative! got {value}",
          stringify!($value)
        ))
        .unwrap();
      return $err;
    }
    value as _
  }};
}

#[macro_export]
macro_rules! component_name {
  ($value: expr) => {{
    const DATA: [u8; 0x30] = {
      let name = { $value };
      if name.len() >= 0x30 {
        panic!("name length must be less than 64 characters")
      }
      let mut new_name = [0; 0x30];
      let mut index = name.len();
      loop {
        index -= 1;
        new_name[index] = name[index];
        if index == 0 {
          break;
        }
      }
      new_name
    };

    DATA.clone()
  }};
}

#[macro_export]
macro_rules! range_of_field {
  ($type: ty, $field: ident) => {{
    const RANGE: std::ops::Range<usize> = {
      let size = {
        let v: std::mem::MaybeUninit<$type> = std::mem::MaybeUninit::uninit();
        size_of_val(unsafe { &v.assume_init_ref().$field })
      };

      let offset = std::mem::offset_of!($type, $field);

      offset as _..(offset + size) as _
    };

    let range: std::ops::Range<_> = RANGE.start as _..RANGE.end as _;
    range
  }};
}

fn overlapping<Idx: PartialOrd>(a: &Range<Idx>, b: &Range<Idx>) -> bool {
  a.contains(&b.start) && a.contains(&b.end)
}

// fn subtract_range(a: &Range<usize>, b: &Range<usize>) -> Range<usize> {
//   a.start - b.start..a.end - b.end
// }

pub fn align_up_to_page<T: PrimInt>(value: T) -> T {
  align_up(value, T::one().unsigned_shl(12))
}

pub fn align_up<T: PrimInt>(value: T, alignment: T) -> T {
  (value + alignment - T::one()) & !(alignment - T::one())
}

#[unsafe(no_mangle)]
extern "system" fn JNI_OnLoad<'local>(_: JNIEnv<'local>, _: usize) -> jint {
  JNI_VERSION_1_8
}
